% 
% Expert System Project: Troubleshooting Computer Failures
%
% Group Members: Particle, Zhang Lishan, Zheng Hui, Li Liting
% 
%  RESTRICTIONS: YOU CAN USE THIS SOFTWARE ANYWHERE AS LONG AS YOU
%                GIVE FULL CREDIT TO ALL GROUP MEMBERS.
% 
%  NOTICE:       GROUP MEMBERS ARE NOT RESPONSIBLE FOR ANYTHING BAD
%                THIS SOFTWARE DOES. IF YOU'RE RUNNING THIS PROGRAM,
%                YOU TAKE ALL THE RESPONSIBILITY UPON YOURSELF.
%
% CONTACT PARTICLE IF YOU HAVE ANY OTHER QUESTIONS (Particle's e-mail:
% is iparticle@yahoo.com or whatever is specified on the website where
% you got this file.)
%
% ABOUT THIS CODE: This code is mostly a modified version of an 
%                  Expert System Shell found in Ivan Bratko's 
%                  Prolog, Programming for Artifical Intelligence, 2nd Ed. 
%
% Program tested using GNU Prolog (Linux)... simply start GNU Prolog,
% type:
% | ?- [expert].
% expert.
%
% and the prompt that asks a question should appear. Then, just
% provide whatever you'd like to verify, and the system will ask you
% questions about it. For example:
%
% Question, plese
% monitor problem.
%
% Is it true: monitor plugged in? 
%
% at which point you asnwer 'yes' or 'no'.
%





%%
%% finds answer to a given goal
%%


:- op(900,xfx,:).
:- op(800,xfx,was).
:- op(870,fx,if).
:- op(880,xfx,then).
:- op(550,xfy,or).
:- op(540,xfy,and).
:- op(300,fx,'derived by').
:- op(600,xfx,from).
:- op(600,xfx,by).
:- op(700,xfx,is).
:- op(500,fx,not).

not(P) :-
	P,!,fail
	;
	true.

explore(Goal,_,Goal is true was 'found as a fact') :-
	fact: Goal.
	
explore(Goal,Trace,Goal is TruthValue was 'derived by' Rule from Answer) :-
	Rule: if Condition then Goal,
	explore(Condition,[Goal by Rule|Trace],Answer),
	truth(Answer,TruthValue).
	
explore(Goal1 and Goal2,Trace,Answer) :- !,
	explore(Goal1,Trace,Answer1),
	continue(Answer1,Goal1 and Goal2,Trace,Answer).
	
explore(Goal1 or Goal2,Trace,Answer) :-
	exploreyes(Goal1,Trace,Answer)
	;
	exploreyes(Goal2,Trace,Answer).
	
explore(Goal1 or Goal2,Trace,Answer1 and Answer2) :- !,
	not exploreyes(Goal1,Trace,_),
	not exploreyes(Goal2,Trace,_),
	explore(Goal1,Trace,Answer1),
	explore(Goal2,Trace,Answer2).
	
explore(Goal,Trace,Goal is Answer was told) :-
	useranswer(Goal,Trace,Answer).
	
exploreyes(Goal,Trace,Answer) :-
	explore(Goal,Trace,Answer),
	positive(Answer).
	
continue(Answer1,_ and Goal2,Trace,Answer) :-
	positive(Answer1),
	explore(Goal2,Trace,Answer2),
	(positive(Answer2),
		Answer = Answer1 and Answer2
		;
		negative(Answer2),
		Answer = Answer2
	).
	
continue(Answer1,_ and _,_,Answer1) :-
	negative(Answer1).
	
truth(_ is TruthValue was _,TruthValue) :- !.

truth(Answer1 and Answer2,TruthValue) :-
	truth(Answer1,true),
	truth(Answer2,true), !,
	TruthValue = true
	;
	TruthValue = false.
	
positive(Answer) :-
	truth(Answer,true).
	
negative(Answer) :-
	truth(Answer,false).
	

%%
%% user answer
%%

useranswer(Goal,Trace,Answer) :-
	askable(Goal,_),
	freshcopy(Goal,Copy),
	useranswer(Goal,Copy,Trace,Answer,1).
	
useranswer(Goal,_,_,_,N) :-
	N > 1,
	instantiated(Goal), !,
	fail.
	
useranswer(Goal,Copy,_,Answer,_) :-
	wastold(Copy,Answer,_),
	instance_of(Copy,Goal), !.
	
useranswer(Goal,_,_,true,N) :-
	wastold(Goal,true,M),
	M >= N.
	
useranswer(Goal,Copy,_,_,_) :-
	end_answers(Copy),
	instance_of(Copy,Goal), !,
	fail.

useranswer(Goal,_,Trace,Answer,N) :-
	askuser(Goal,Trace,Answer,N).
	
askuser(Goal,Trace,Answer,N) :-
	askable(Goal,ExternFormat),
	format(Goal,ExternFormat,Question,[],Variables),
	ask(Goal,Question,Variables,Trace,Answer,N).
	
ask(Goal,Question,Variables,Trace,Answer,N) :-
	nl,
	(Variables = [], !,
		write('Is it true: ')
		;
		write('Any (more) solutions to: ')
	),
	write(Question),write('? '),
	getreply(Reply),!,
	process(Reply,Goal,Question,Variables,Trace,Answer,N).
	
process(why,Goal,Question,Variables,Trace,Answer,N) :-
	showtrace(Trace),
	ask(Goal,Question,Variables,Trace,Answer,N).
	
process(yes,Goal,_,Variables,Trace,true,_) :-
	nextindex(Next),
	Next1 is Next + 1,
	(askvars(Variables),
		assertz(wastold(Goal,true,Next))
		;
		freshcopy(Goal,Copy),
		useranswer(Goal,Copy,Trace,_,Next1)
	).
	
process(no,Goal,_,_,_,false,_) :-	
	freshcopy(Goal,Copy),
	wastold(Copy,true,_), !,
	assertz(end_answers(Goal)),
	fail
	;
	nextindex(Next),
	assertz(wastold(Goal,false,Next)).
	
format(Var,Name,Name,Vars,[Var/Name|Vars]) :-
	var(Var), !.
	
format(Atom,Name,Atom,Vars,Vars) :-
	atomic(Atom), !,
	atomic(Name).
	
format(Goal,Form,Question,Vars0,Vars) :-
	Goal =.. [Functor | Args1],
	Form =.. [Functor | Forms],
	formatall(Args1,Forms,Args2,Vars0,Vars),
	Question =.. [Functor | Args2].
	
formatall([],[],[],Vars,Vars).

formatall([X|XL],[F|FL],[Q|QL],Vars0,Vars) :-
	formatall(XL,FL,QL,Vars0,Vars1),
	format(X,F,Q,Vars1,Vars).
	
askvars([]).

askvars([Variable/Name | Variables]) :-
	nl, write(Name), write(' = '),
	read(Variable),
	askvars(Variables).
	
showtrace([]) :-
	nl, write('This was your question '), nl.
	
showtrace([Goal by Rule | Trace]) :-
	nl, write('To investigate by '),
	write(Rule),write(','),
	write(Goal),
	showtrace(Trace).
	
instantiated(Term) :-
	numbervars(Term,0,0).

instance_of(Term,Term1) :-
	freshcopy(Term1,Term2),
	numbervars(Term2,0,_),!,
	Term = Term2.
	
freshcopy(Term,FreshTerm) :-
	asserta(copy(Term)),
	retract(copy(FreshTerm)),!.
	

%% GNU Prolog rework; see "expert" rule.
%% lastindex(0).

nextindex(Next) :-
	retract(lastindex(Last)),!,
	Next is Last + 1,
	asserta(lastindex(Next)).
	


%%
%% how explanation
%%

present(Answer) :-
	nl,showconclusion(Answer),
	nl,write('Would you like to see how? '),
	getreply(Reply),
	(Reply = yes, !,
		show(Answer)
		;
		true
	).
	
showconclusion(Answer1 and Answer2) :- !,
	showconclusion(Answer1),write(' and '),
	showconclusion(Answer2).
	
showconclusion(Conclusion was _) :-
	write(Conclusion).
	
show(Solution) :-
	nl,show(Solution,0),!.
	
show(Answer1 and Answer2,H) :- !,
	show(Answer1,H),
	tab(H),write(and),nl,
	show(Answer2,H).
	
show(Answer was Found,H) :-
	tab(H),writeans(Answer),
	nl,tab(H),
	write('was '),
	show1(Found,H).
	
show1(Derived from Answer,H) :- !,
	write(Derived),write(' from '),
	nl,H1 is H + 4,
	show(Answer,H1).
	
show1(Found,_) :-
	write(Found),nl.
	
writeans(Goal is true) :- !,
	write(Goal).
	
writeans(Answer) :-
	write(Answer).
	
expert :-

%% this is a fix for GNU Prolog, since it doesn't allow modification of 
%% static rules (as Bratko's Book assumes) (this was driving me nuts for
%% hours!)
	asserta(lastindex(0)),

%% another GNU Prolog rework (a rule has to be defined, even if nothing
%% that matches is found for that rule.)
	assertz(wastold(a,b,c)),

%% same thing as above (rule needs to be defined)
	assertz(end_answers(a)),

	getquestion(Question),
	( answeryes(Question)
		;
		answerno(Question)
	).

answeryes(Question) :-
	markstatus(negative),
	explore(Question,[],Answer),
	positive(Answer),
	markstatus(positive),
	present(Answer),nl,
	write('More solutions? '),
	getreply(Reply),
	Reply = no.
	
answerno(Question) :-
	retract(no_positive_answer_yet),!,
	explore(Question,[],Answer),
	negative(Answer),
	present(Answer),nl,
	write('More negative solutions? '),
	getreply(Reply),
	Reply = no.
	
markstatus(negative) :-
	assertz(no_positive_answer_yet).
	
markstatus(positive) :-
	retract(no_positive_answer_yet),!
	;
	true.
	
getquestion(Question) :-
	nl,write('Question, plese'),nl,
	read(Question).
	
getreply(Reply) :-
	read(Answer),
	means(Answer,Reply), !
	;
	nl,write('Answer unknown, try again plese'),nl,
	getreply(Reply).
	
means(yes,yes).
means(y,yes).
means(no,no).
means(n,no).
means(why,why).
means(w,why).










:- op(100,xfx,[looks,feels,smells,isa,'connected to',has]).
:- op(100,xf,[problem,'not working','plugged in','turned on']).
:- op(100,fx,[visual_prob,comm_prob,input_prob]).
:- op(100,fx,[output_prob,monitor_prob,videocard_prob]).
:- op(100,fx,[modem_prob,network_prob,mouse_prob,keyboard_prob]).
:- op(100,xfx,['works in','works with','does not',generates]).
:- op(100,fx,[connected,sound_prob,soundcard_prob,speaker_prob]).
:- op(100,fx,[storage_prob,harddrive_prob,cdrom_prob]).
:- op(100,xfx,[sounds]).

%% see if the component has a problem.
rule1: if
	visual_prob Component
	or
	storage_prob Component
	or
	comm_prob Component
	or
	output_prob Component
	or
	input_prob Component
	then
	Component problem.

%%%% the below rules need to be modified to include
%%%% actual problem symptoms...

% see if the component is a visual component that has a problem.
rule2: if
	Component isa visual_dev
	and
	(monitor_prob Component
	or
	videocard_prob Component)
	then
	visual_prob Component.

% see if the component is a communications component and has a problem.
rule3: if
	Component isa comm_dev
	and
	(modem_prob Component
	or
	network_prob Component)
	then
	comm_prob Component.

% see if the component is an input device and has a problem.
rule4: if
	Component isa input_dev
	and
	(mouse_prob Component
	or
	keyboard_prob Component) 
	then
	input_prob Component.

% see if the component is an output device, and has a problem
rule5: if
	Component isa output_dev
	and
	(visual_prob Component
	or
	comm_prob Component
	or
	sound_prob Component)
	then
	output_prob Component.

% if we're dealing with a monitor type of device, determine
% if it's connected, and then see if it has any problems.
rule6:  if
	 Component isa monitor
	and
	connected Component
	and
	(Component looks black
	or
	Component looks flickery
	or
	Component looks blury)
	and
	Component 'does not' 'work with another computer'	
	then
	monitor_prob Component.

% if the component is a visual device has has problems.
rule7: if
	 Component isa videocard
	and
	(Component generates 'terrible looking images'
	or
	Component generates 'blury images'
	or
	Component feels slow)
	and
	Component 'does not' 'work inside another computer'
	and
	Component has 'all drivers installed'
	then
	videocard_prob Component.	

% determine whether a component is fully connected.
rule8: if
	Component 'plugged in'
	and
	Component 'turned on'
	and
	Component 'connected to' computer
	then
	connected Component.

% network cards
rule9:
	if
	Component isa network_card
	and
	(Component feels 'extremely slow'
	or
	Component looks dead)
	and
	Component has 'all drivers installed'
	and
	Component 'does not' 'work inside another computer'
	then
	network_prob Component.

% modems
rule10:
	if
	Component isa modem
	and
	(Component 'does not' 'dial out'
	or
	Component 'does not' 'connect'
	or
	Component 'does not' 'answer the phone')
	and
	Component has 'all drivers installed'
	and
	Component 'does not' 'work inside another computer'	
	then
	modem_prob Component.

% the mouse
rule11: if
	  Component isa mouse
	and
	Component 'connected to' computer
	and
	Component 'does not' 'respond when moved'
	and
	Component 'does not' 'respond when moved AFTER a reboot'
	then
	mouse_prob Component.

% the keyboard
rule12: if
	 Component isa keyboard
	and
	Component 'connected to' computer
	and
	(Component 'does not' 'respond when pressed'
	or
	Component generates 'a beeping sound every time you type a key')
	then
	keyboard_prob Component.

% sound problem
rule13: if
	 soundcard_prob Component
	or
	 speaker_prob Component
	then
	sound_prob Component.

% it's the speakers!
rule14: if
	 Component isa speaker
	and
	connected Component
	and
	(Component generates 'no sound'
	or
	Component generates 'cracking sound')
	then
	speaker_prob Component.

% it's the sound card itself
rule15: if
	 Component isa soundcard
	 and
	 Component has 'all drivers installed'
	 and
	 (Component generates 'no sound'
	 or
	 Component generates 'cracking sound'
	 or
	 Component has 'lots of noise')
	then
	soundcard_prob Component.

rule16: if
	Component isa storage_dev
	and
	(harddrive_prob Component
	 or
	 cdrom_prob Component)		
	then
	storage_prob Component.
	
rule17: if
	 Component isa harddrive
	 and
	( Component sounds noisy 
	  or
	  Component 'does not' 'save data' )
	then 
	harddrive_prob Component.
rule18: if
	 Component isa cdrom
	 and
	 (Component 'does not' 'read cds'
	  or
	  Component sounds noisy )
	then
	cdrom_prob Component.

fact: X isa visual_dev :-
	member(X,[monitor,screen,display,videocard,video]).
fact: X isa comm_dev :-
	member(X,[modem,isdn,network]).
fact: X isa input_dev :-
	member(X,[keyboard,mouse,joystick,scanner,modem]).
fact: X isa output_dev :-
	member(X,[screen,speakers,modem]).
fact: X isa storage_dev :-
	member(X,[harddrive,cdrom,floppy,zipdrive]).

fact: X isa monitor :-
	member(X,[monitor,flatscreen]).
fact: X isa videocard :-
	member(X,[videocard,accelerator]).
fact: X isa modem :-
	member(X,[modem,serial]).
fact: X isa network_card :-
	member(X,[network,network_card,lan]).
fact: X isa keyboard :-
	member(X,[keyboard]).
fact: X isa mouse :-
	member(X,[mouse]).
fact: X isa soundcard :-
	member(X,[soundcard]).
fact: X isa speaker :-
	member(X,[speaker,speakers]).
fact: X isa harddrive :-
	member(X,[harddrive,hd]).
fact: X isa cdrom :-
	member(X,[cdrom]). 

askable(_ looks _,'Component' looks 'LikeWhat').
askable(_ feels _,'Component' feels 'LikeWhat').
askable(_ smells _,'Component' smells 'LikeWhat').
askable(_ 'not working','Component' 'not working').
askable(_ 'connected to' _,'Component' 'connected to' 'What').
askable(_ 'plugged in','Component' 'plugged in').
askable(_ 'turned on','Component' 'turned on').
askable(_ 'works in' _,'Component' 'works in' 'What').
askable(_ 'works with' _,'Component' 'works with' 'What').
askable(_ 'does not' _,'Component' 'does not' 'DoWhat').
askable(_ generates _,'Component' generates 'What').
askable(_ has _,'Component' has 'What').
askable(_ sounds _,'Component' sounds 'LikeWhat').
